#!/usr/bin/env python3

# TODO: 处理颜色嵌套后，导致色彩控制字符串冗余的问题

import sys
from pathlib import Path
import logging

from logging import LogRecord
from copy import deepcopy

VERSION = '0.0.10'
PKG_ROOT = Path(__file__).parent


def white(text): return '\x1B[1;37m' + text + '\x1B[0m'
def blue(text): return '\x1B[1;34m' + text + '\x1B[0m'
def green(text): return '\x1B[1;32m' + text + '\x1B[0m'
def yellow(text): return '\x1B[1;33m' + text + '\x1B[0m'
def red(text): return '\x1B[1;31m' + text + '\x1B[0m'
def white_red(text): return '\x1B[47;31m' + text + '\x1B[0m'

def readable_hex(data): return ' '.join('{:02x}'.format*(b) for b in data)

LEVELNAME_PREFIX = '['
LEVELNAME_SUFFIX = ']'

OK = green("OK").join([LEVELNAME_PREFIX, LEVELNAME_SUFFIX])
DEBUG = white("DEBUG").join([LEVELNAME_PREFIX, LEVELNAME_SUFFIX])
INFO = blue("INFO").join([LEVELNAME_PREFIX, LEVELNAME_SUFFIX])
WARNING = yellow("WARNING").join([LEVELNAME_PREFIX, LEVELNAME_SUFFIX])
ERROR = red("ERROR").join([LEVELNAME_PREFIX, LEVELNAME_SUFFIX])
CRITICAL = white_red("CRITICAL").join([LEVELNAME_PREFIX, LEVELNAME_SUFFIX])

DEBUG_INDENT    = " "*8
INFO_INDENT     = " "*7
WARNING_INDENT  = " "*10
ERROR_INDENT    = " "*8
CRITICAL_INDENT = " "*11


class Formatter(logging.Formatter):
    colorful_levelname = {
        logging.DEBUG:    DEBUG,
        logging.INFO:     INFO,
        logging.WARNING:  WARNING,
        logging.ERROR:    ERROR,
        logging.CRITICAL: CRITICAL
    }
    
    level_indent = {
        logging.DEBUG:    DEBUG_INDENT,
        logging.INFO:     INFO_INDENT,
        logging.WARNING:  WARNING_INDENT,
        logging.ERROR:    ERROR_INDENT,
        logging.CRITICAL: CRITICAL_INDENT
    }

    def __init__(self, msg, colorful: bool = True, auto_indent: bool = True):
        super().__init__(msg)
        self.colorful = colorful
        self.auto_indent = auto_indent
        # print(colorful)

    def format(self, record: LogRecord):
        """
        record - 如果直接修改 LogRecord 实例而不是修改它的复制品，就会影响后续的处理。因为
                 LogRecord 实例会继续传递给下一个 logging.Handler 实例，并被该 handler 
                 实例设置的 formatter 处理。
        """
        record = deepcopy(record)

        if self.colorful:
            try:
                record.levelname = self.colorful_levelname[record.levelno]
            except KeyError:
                record.levelname = record.levelname.join(
                    [LEVELNAME_PREFIX, LEVELNAME_SUFFIX])
        else:
            record.levelname = record.levelname.join(
                [LEVELNAME_PREFIX, LEVELNAME_SUFFIX])

        if self.auto_indent:
            try:
                # record.msg 不一定是 str，比如调用 logger.debug(1234) 时，record.msg
                # 类型就是 int。
                if isinstance(record.msg, str):
                    record.msg = record.msg.replace('\n', '\n'+self.level_indent[record.levelno])
            except KeyError:
                # unknown log level
                pass
        
        if record.levelno == logging.DEBUG or record.levelno == logging.ERROR:
            if self.auto_indent:
                # 如果 level 是 debug 或 error，那么我们为 msg 添加一个 name 前缀。目的是
                # 方便定位 logger.debug 或 logger.error 在源码中的输出位置。
                # 此时 log 的输出样子如下：
                #     logger.debug("test")
                #     [DEBUG] name
                #             test
                record.msg = record.name + '\n' + self.level_indent[record.levelno] + \
                             record.msg
            else:
                record.msg = 'From "{}". {}'.format(record.name, record.msg)

        return super().format(record)


class Logger(logging.getLoggerClass()):
    """Inherit the standard Logger class, or the last class passed to 
    setLoggerClass() to ensure this customized Logger class will not undo 
    customizations already applied by other code.
    """

    def __init__(self, name, level: int, filename: str = None, 
                 colorful: bool = True, auto_indent: bool = True):
        """
        auto_indent - log 在输出的时候会有表示 level 的头部信息，但是当 log 换行的时候，
                      新行并不会自动缩进，从而对齐到头部信息末尾的列再输出。这一定程度上不
                      是很美观。使能该 flag 可以让上述情况的新行自动缩进。比如：
                      
                      ```txt
                      使能前
                      logger.info("Message\nnewline")
                      [INFO] Message
                      new line
                      
                      使能后
                      logger.info("Message\nnewline")
                      [INFO] Message
                             new line
                      ```           
        """
        super().__init__(name, level)

        # 设置 Formatter，使最终打印到 stdout 的内容由两个部分组成：
        #     第一个部分是 LogRecord.levelname；
        #     第二个部分是 LogRecord.message。
        # 该 formatter 在 log 输出到 stdout 时使用。
        sh_formatter = Formatter('%(levelname)s %(message)s', colorful=colorful, 
                                 auto_indent=auto_indent)
        # 该 formatter 在 log 输出到 file 时使用。
        fh_formatter = Formatter('%(levelname)s %(message)s', colorful=False, 
                                 auto_indent=auto_indent)

        # 对 log 输出到 stdout 的情况做基本配置。
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(level)
        sh.setFormatter(sh_formatter)
        self.addHandler(sh)

        # 对 log 输出到 file 的情况做基本配置。
        if filename is not None:
            fh = logging.FileHandler(filename)
            fh.setLevel(level)
            fh.setFormatter(fh_formatter)
            self.addHandler(fh)


def __test():
    logger = Logger(__name__, logging.NOTSET, 'log', auto_indent=True)
    logger.log(1, 'asdf\nasdfs')
    logger.debug('asdfs %d\nasdfssadfasdfas\ndasasdf', 1)
    logger.debug(21341234)
    logger.info('I told you so %d\nkjlhkljhjkjkh', 1)
    logger.warning('Watch out!')
    logger.error("an error occurrence！\nsdfasfd")
    logger.critical('asdf')


if __name__ == '__main__':
    __test()
